from __future__ import annotations
import asyncio
import json
import logging
import re
import sys
import types
import inspect
import uuid
from enum import Enum
from typing import (
    Any,
    Optional,
    Type,
    List,
    Dict,
    Tuple,
    get_args,
    Awaitable,
    no_type_check_decorator,
    Union,
    get_origin,
    Protocol,
)
from dbus_next.aio import MessageBus
import dbus_next.service

from pyobs.comm import Comm
from pyobs.events import ModuleOpenedEvent, ModuleClosedEvent, Event
from pyobs.events.event import EventFactory
from pyobs.interfaces import Interface
from pyobs.utils.parallel import Future

log = logging.getLogger(__name__)


NONE_VALUES = {str: "NONE", int: sys.maxsize, float: sys.maxsize, list: ["EMPTY"]}


class ServiceInterface(dbus_next.service.ServiceInterface):
    def __init__(self, interface: str, comm: DbusComm, pyobs_interfaces: List[str]):
        dbus_next.service.ServiceInterface.__init__(self, interface)
        self._interfaces = pyobs_interfaces
        self._comm = comm

    @no_type_check_decorator
    @dbus_next.service.method()
    def get_interfaces(self) -> "as":
        return self._interfaces

    @no_type_check_decorator
    @dbus_next.service.method(sender_keyword="sender")
    async def handle_event(self, event: "s", sender):
        # convert event to dict
        try:
            d = json.loads(event.replace("'", '"'))
        except json.decoder.JSONDecodeError:
            return

        # create event
        ev = EventFactory.from_dict(d)

        # get real sender
        real_sender = await self._comm.get_dbus_owner(sender)

        # handle it
        await self._comm.handle_event(ev, real_sender)

    @no_type_check_decorator
    @dbus_next.service.method()
    def set_timeout(self, uid: "s", timeout: "d"):
        self._comm.set_timeout(uid, timeout)


class DbusMethod(Protocol):
    def __call__(self, *args: Any) -> Any:
        ...


class DbusComm(Comm):
    """A Comm class using cbus.

    This Comm class uses dbus for communication between modules and is therefore available on all (most) Linux
    systems. The interface name for the bus is build as <domain>.<name> and this class will only find other modules
    with the same domain, so it should be kept constant in a closed system. The name on the other hand should be
    unique in the system.

    A basic configuration looks like this::

        comm:
            class: pyobs.dbus.DbusComm
            name: example
    """

    __module__ = "pyobs.comm.dbus"

    def __init__(
        self,
        name: str,
        domain: str = "org.pyobs.module",
        *args: Any,
        **kwargs: Any,
    ):
        """Create a new dbus Comm module.

        Args:
            name: Name for export.
            domain: Domain for export.
        """
        Comm.__init__(self, *args, **kwargs)

        # variables
        self._name = name
        self._domain = domain
        self._dbus: Optional[MessageBus] = None
        self._dbus_classes: Dict[str, dbus_next.service.ServiceInterface] = {}
        self._dbus_introspection: Optional[dbus_next.aio.ProxyInterface] = None
        self._dbus_introspection_cache: Dict[str, dbus_next.introspection.Node] = {}
        self._interfaces: Dict[str, List[Type[Interface]]] = {}
        self._futures: Dict[str, Future] = {}

    async def open(self) -> None:
        """Creates the dbus connection."""

        # create client
        self._dbus = await MessageBus().connect()
        if not self._dbus:
            raise ValueError("Could not create DBUS connection.")

        # build and publish classes
        self._build_dbus_classes()
        for interface, obj in self._dbus_classes.items():
            self._dbus.export("/" + interface.replace(".", "/"), obj)
            await self._dbus.request_name(interface)

        # get object for introspection
        introspection = await self._dbus.introspect("org.freedesktop.DBus", "/org/freedesktop/DBus")
        proxy_object = self._dbus.get_proxy_object("org.freedesktop.DBus", "/org/freedesktop/DBus", introspection)
        if not proxy_object:
            raise ValueError("Could not fetch proxy object for DBUS.")
        self._dbus_introspection = proxy_object.get_interface("org.freedesktop.DBus")

        # get client list
        await self._update_client_list()

        # update task
        asyncio.create_task(self._update())

        # open Comm
        await Comm.open(self)

    async def close(self) -> None:
        """Close connection."""

        # close parent class
        await Comm.close(self)

        # disconnect from dbus
        if self._dbus:
            self._dbus.disconnect()

    async def _update(self) -> None:
        """Periodic updates."""
        while True:
            await self._update_client_list()
            await asyncio.sleep(5)

    async def _update_client_list(self) -> None:
        """Update list of clients."""

        # check
        if self._dbus_introspection is None:
            return

        # get all interfaces containing "pyobs"
        x = await self._dbus_introspection.call_list_names()
        data = list(filter(lambda d: self._domain in d, x))

        # get all modules: first run regexp on all entries and then cut by length of prefix
        prefix = self._domain + "."
        r = re.compile(prefix + r"(\w+)$")
        modules = list(map(lambda d: d[len(prefix) :], filter(r.match, data)))

        # get interfaces
        interfaces: Dict[str, List[Type[Interface]]] = {}
        for m in modules:
            prefix = f"{self._domain}.{m}.interfaces."
            iface_names = list(map(lambda d: d[len(prefix) :], filter(lambda d: d.startswith(prefix), data)))
            interfaces[m] = self._interface_names_to_classes(iface_names)

        # loop newly connected modules except myself
        for m in list(set(interfaces.keys()) - set(self._interfaces.keys()) - {self._name}):
            # send event
            self._send_event_to_module(ModuleOpenedEvent(), m)

        # loop freshly disconnected modules except myself
        for m in list(set(self._interfaces.keys()) - set(interfaces.keys()) - {self._name}):
            # send event
            self._send_event_to_module(ModuleClosedEvent(), m)

        # store interfaces
        self._interfaces = interfaces

    def _annotation_to_dbus(self, annotation: Any) -> Any:
        if hasattr(annotation, "__origin__") and annotation.__origin__ == list:
            # lists
            return "a" + self._annotation_to_dbus(get_args(annotation)[0])
        elif hasattr(annotation, "__origin__") and annotation.__origin__ == tuple:
            # tuples
            return "(" + "".join([self._annotation_to_dbus(a) for a in get_args(annotation)]) + ")"
        elif hasattr(annotation, "__origin__") and annotation.__origin__ == dict:
            # dicts
            return "a{" + "".join([self._annotation_to_dbus(a) for a in get_args(annotation)]) + "}"
        elif (
            hasattr(annotation, "__origin__") and annotation.__origin__ is Union and type(None) in get_args(annotation)
        ):
            # optional parameter
            typ = list(filter(lambda x: x is not None, get_args(annotation)))[0]
            return self._annotation_to_dbus(typ)
        elif inspect.isclass(annotation) and issubclass(annotation, Enum):
            return "s"
        else:
            try:
                return {int: "i", float: "d", str: "s", bool: "b", Any: "s"}[annotation]
            except KeyError:
                raise

    def _build_dbus_signature(self, sig: inspect.Signature) -> inspect.Signature:
        # build list of parameters
        new_params: List[inspect.Parameter] = []

        # real parameters
        for name, param in sig.parameters.items():
            # ignore kwargs
            if name == "kwargs":
                continue

            # get new annotation
            annotation = param.annotation if name == "self" else self._annotation_to_dbus(param.annotation)

            # build new param
            p = inspect.Parameter(
                name,
                kind=inspect.Parameter.POSITIONAL_ONLY if name == "self" else inspect.Parameter.POSITIONAL_OR_KEYWORD,
                annotation=annotation,
            )

            # store it
            new_params.append(p)

        # add "s" at the end for uid
        new_params.append(inspect.Parameter("uid", kind=inspect.Parameter.KEYWORD_ONLY, annotation="s"))

        # return type
        return_annotation = None if sig.return_annotation is None else self._annotation_to_dbus(sig.return_annotation)

        # create new signature
        return inspect.Signature(parameters=new_params, return_annotation=return_annotation)

    def _build_dbus_classes(self) -> None:
        # got a module?
        if self._module is not None:
            # main module
            main_klass = types.new_class(self._name, bases=(ServiceInterface,))

            # loop all interfaces features
            for i in self._module.interfaces:
                # create class
                klass_name = f"{self._name}_{i.__name__}"
                interface = f"{self._domain}.{self._name}.interfaces.{i.__name__}"
                klass = types.new_class(klass_name, bases=(dbus_next.service.ServiceInterface,))

                # loop all methods:
                for func_name, func in inspect.getmembers(i, predicate=inspect.isfunction):
                    # get signature
                    sig = inspect.signature(func)

                    # set method
                    my_func = types.MethodType(self._dbus_function_wrapper(func_name, sig), self)
                    setattr(main_klass, func_name, my_func)

                # initialize it
                self._dbus_classes[interface] = klass(interface)

            # initialize main class
            interface = f"{self._domain}.{self._name}"
            self._dbus_classes[interface] = main_klass(interface, self, [i.__name__ for i in self._module.interfaces])

    async def _get_dbus_introspection(self, client: str) -> Tuple[dbus_next.introspection.Node, str, str]:
        # check
        if self._dbus is None:
            raise ValueError("No connection")

        # get interface and path
        interface = f"{self._domain}.{client}"
        path = "/" + interface.replace(".", "/")

        # get introspection and return it
        if client not in self._dbus_introspection_cache:
            self._dbus_introspection_cache[client] = await self._dbus.introspect(interface, path)
        return self._dbus_introspection_cache[client], interface, path

    async def _get_dbus_method(self, client: str, method: str) -> DbusMethod:
        # check
        if self._dbus is None:
            raise ValueError("No connection")

        # get introspection, interface and path
        introspection, interface, path = await self._get_dbus_introspection(client)

        # get object and module
        obj = self._dbus.get_proxy_object(interface, path, introspection)
        module = obj.get_interface(interface)

        # get function
        return getattr(module, f"call_{method}")

    def _dbus_function_wrapper(self, method: str, sig: inspect.Signature) -> Any:
        """Function wrapper for dbus methods.

        Args:
            method: Name of method to wrap.

        Returns:
            Wrapper.
        """

        async def inner(this: Any, *args: Any, **kwargs: Any) -> Any:
            # check
            if self.module is None:
                return

            # get sender
            sender = None
            if "sender" in kwargs:
                # get real sender
                sender = await self.get_dbus_owner(kwargs["sender"])
                del kwargs["sender"]

            # get uid
            uid = args[-1]
            args = args[:-1]

            # get method
            func, signature, _ = self._module.methods[method]

            # bind parameters
            ba = signature.bind(*args)
            ba.apply_defaults()

            # do we have a timeout?
            if hasattr(func, "timeout") and sender is not None:
                timeout = await getattr(func, "timeout")(self._module, **ba.arguments)
                if timeout:
                    # get introspection, proxy and interface
                    set_timeout = await self._get_dbus_method(sender, "set_timeout")
                    await set_timeout(uid, int(timeout))

            # call method
            return await self.module.execute(method, *args, sender=sender)

        # set signature
        inner.__signature__ = self._build_dbus_signature(sig)
        # TODO: Nicer way to do this?
        inner.__dict__["__DBUS_METHOD"] = dbus_next.service._Method(
            inner, method, disabled=False, sender_keyword="sender"
        )
        return inner
        # return dbus_next.service.method(name=method)(inner)

    async def get_dbus_owner(self, bus: str, attempts: int = 3) -> str:
        """Gets the owning module name for a given bus.

        Params:
            bus: Name of bus to find owner for.

        Returns:
            Owning module.
        """

        # check
        if self._dbus_introspection is None:
            raise ValueError("No connection.")

        # loop all clients, get their owner, and check against bus
        for c in self.clients:
            try:
                owner = await self._dbus_introspection.call_get_name_owner(f"{self._domain}.{c}")
            except dbus_next.errors.DBusError:
                break
            if owner == bus:
                return c

        # nothing found
        if attempts > 0:
            await asyncio.sleep(0.5)
            await self._update_client_list()
            return await self.get_dbus_owner(bus, attempts - 1)
        else:
            raise ValueError("Owner not found.")

    @property
    def name(self) -> Optional[str]:
        """Name of this client."""
        return self._name

    @property
    def clients(self) -> List[str]:
        """Returns list of currently connected clients.

        Returns:
            (list) List of currently connected clients.
        """

        # return names
        return list(self._interfaces.keys())

    async def get_interfaces(self, client: str) -> List[Type[Interface]]:
        """Returns list of interfaces for given client.

        Args:
            client: Name of client.

        Returns:
            List of supported interfaces.

        Raises:
            IndexError: If client cannot be found.
        """

        # return list
        return self._interfaces[client]

    async def _supports_interface(self, client: str, interface: Type[Interface]) -> bool:
        """Checks, whether the given client supports the given interface.

        Args:
            client: Client to check.
            interface: Interface to check.

        Returns:
            Whether or not interface is supported.
        """

        # does it exist?
        return interface in self._interfaces[client]

    async def execute(self, client: str, method: str, annotation: Dict[str, Any], *args: Any) -> Any:
        """Execute a given method on a remote client.

        Args:
            client (str): ID of client.
            method (str): Method to call.
            annotation: Method annotation.
            *args: List of parameters for given method.

        Returns:
            Passes through return from method call.
        """

        # check
        if self._dbus is None:
            raise ValueError("No connection.")

        # create a future for this
        uid = str(uuid.uuid4())
        future = Future(annotation=annotation, comm=self)
        self._futures[uid] = future

        # get method
        func = await self._get_dbus_method(client, method)

        # get method and call it in background
        asyncio.create_task(self._wait_for_method(func(*args, uid), uid))

        # don't wait for response, just return future
        return await future

    async def _wait_for_method(self, func: Awaitable[Any], uid: str) -> None:
        # wait for result
        result = await func

        # set result
        self._futures[uid].set_result(result)

    def cast_to_simple_pre(self, value: Any, annotation: Optional[Any] = None) -> Tuple[bool, Any]:
        """Special treatment of single parameters when converting them to be sent via Comm.

        Args:
            value: Value to be treated.
            annotation: Annotation for value.

        Returns:
            A tuple containing a tuple that indicates whether this value should be further processed and a new value.
        """

        if annotation == Any:
            # cast Anys to string
            return True, str(value)

        elif value is None and get_origin(annotation) == Union:
            # get types that are not None
            typs = list(filter(lambda x: x is not None, get_args(annotation)))

            # loop them
            for typ in typs:
                if typ in NONE_VALUES:
                    return True, NONE_VALUES[typ]
                elif get_origin(typ) in NONE_VALUES:
                    return True, NONE_VALUES[get_origin(typ)]
                else:
                    return True, value
            else:
                return True, value

        elif isinstance(value, tuple):
            return False, list(value)

        else:
            return False, value

    def cast_to_real_pre(self, value: Any, annotation: Optional[Any] = None) -> Tuple[bool, Any]:
        """Special treatment of single parameters when converting them after being sent via Comm.

        Args:
            value: Value to be treated.
            annotation: Annotation for value.

        Returns:
            A tuple containing a tuple that indicates whether this value should be further processed and a new value.
        """

        if annotation == Any:
            # try to guess type
            try:
                f = float(value)
                i = int(f)
                return True, i if f == i and "." not in value else f
            except ValueError:
                return True, value
        elif value in NONE_VALUES.values():
            return True, None
        else:
            return False, value

    async def send_event(self, event: Event) -> None:
        """Send an event to other clients.

        Args:
            event (Event): Event to send
        """

        # check
        if self._dbus is None:
            raise ValueError("No connection.")

        # loop all clients
        for client in self.clients:
            # skip myself
            if client == self._name:
                continue

            # get function
            try:
                func = await self._get_dbus_method(client, "handle_event")
            except ValueError:
                # client offline?
                continue

            # call it
            await func(json.dumps(event.to_json()))

    async def handle_event(self, event: Event, sender: str) -> None:
        """Handle event localy, i.e. send it to module.

        Args:
            event: Event to handle.
            sender: Sender of event.
        """
        self._send_event_to_module(event, sender)

    def set_timeout(self, uid: str, timeout: float) -> None:
        """Set timeout, usually received from other module.

        Args:
            uid: UID of remote call.
            timeout: Timeout in seconds.
        """
        self._futures[uid].set_timeout(timeout)


__all__ = ["DbusComm"]



from __future__ import annotations
from typing import List, Optional, Dict, Any, TYPE_CHECKING

from pyobs.utils.enums import ImageType
from pyobs.utils.time import Time

if TYPE_CHECKING:
    from pyobs.images import Image


class FrameInfo:
    """Base class for frame infos."""

    def __init__(self) -> None:
        self.id = None
        self.filename = None
        self.filter_name = None
        self.binning = None
        self.dateobs = None


class Archive:
    """Base class for image archives."""

    __module__ = "pyobs.utils.archive"

    async def list_options(
        self,
        start: Optional[Time] = None,
        end: Optional[Time] = None,
        night: Optional[str] = None,
        site: Optional[str] = None,
        telescope: Optional[str] = None,
        instrument: Optional[str] = None,
        image_type: Optional[ImageType] = None,
        binning: Optional[str] = None,
        filter_name: Optional[str] = None,
        rlevel: Optional[int] = None,
    ) -> Dict[str, List[Any]]:
        """Returns a list of options restricted to the given parameters.

        Args:
            start: Start time for restriction.
            end: End time for restriction.
            night: Images in given night.
            site: From given site.
            telescope: From given telescope.
            instrument: From given instrument.
            image_type: With given image type.
            binning: With given binning.
            filter_name: With given filter.
            rlevel: In given reduction level.

        Returns:
            Dictionary with lists of "binnings", "filters", "imagetypes", "instruments", "sites", and "telescopes".
        """
        raise NotImplementedError

    async def list_frames(
        self,
        start: Optional[Time] = None,
        end: Optional[Time] = None,
        night: Optional[str] = None,
        site: Optional[str] = None,
        telescope: Optional[str] = None,
        instrument: Optional[str] = None,
        image_type: Optional[ImageType] = None,
        binning: Optional[str] = None,
        filter_name: Optional[str] = None,
        rlevel: Optional[int] = None,
    ) -> List[FrameInfo]:
        """Returns a list of frames restricted to the given parameters.

        Args:
            start: Start time for restriction.
            end: End time for restriction.
            night: Images in given night.
            site: From given site.
            telescope: From given telescope.
            instrument: From given instrument.
            image_type: With given image type.
            binning: With given binning.
            filter_name: With given filter.
            rlevel: In given reduction level.

        Returns:
            List of frames.
        """
        raise NotImplementedError

    async def download_frames(self, frames: List[FrameInfo]) -> List[Image]:
        """Download given frames.

        Args:
            frames: List of frames to download.

        Returns:
            List of Image objects.
        """
        raise NotImplementedError

    async def upload_frames(self, frames: List[Image]) -> None:
        raise NotImplementedError


__all__ = ["FrameInfo", "Archive"]



import glob
from pathlib import Path
from typing import List, Dict, Any, Optional
import logging

import pandas as pd
from astropy.io import fits

from pyobs.utils.time import Time
from pyobs.images import Image
from pyobs.utils.archive import Archive, FrameInfo
from pyobs.utils.enums import ImageType

log = logging.getLogger(__name__)


class LocalArchive(Archive):
    """Connector class to a local image archive."""

    __module__ = "pyobs.utils.archive"

    def __init__(self, root: str, **kwargs: Any):
        self._root = Path(root)
        self._data = pd.DataFrame()
        self._update_root()

    def _update_root(self):
        """Update files in root directory."""

        # init lists
        filenames = sorted(glob.glob(str(self._root / "*.fits")))
        columns = {
            h: []
            for h in [
                "date-obs",
                "day-obs",
                "binning",
                "filter",
                "image_type",
                "instrument",
                "site",
                "telescope",
                "rlevel",
            ]
        }

        # loop files
        for filename in filenames:
            # load header
            hdr = fits.getheader(filename)

            # store it
            columns["date-obs"].append(Time(hdr["DATE-OBS"]) if "DATE-OBS" in hdr else None)
            columns["day-obs"].append(Time(hdr["DAY-OBS"]) if "DAY-OBS" in hdr else None)
            columns["binning"].append(f"{hdr['XBINNING']}x{hdr['YBINNING']}" if "XBINNING" in hdr else None)
            columns["filter"].append(hdr["FILTER"] if "FILTER" in hdr else None)
            columns["image_type"].append(hdr["IMAGETYP"] if "IMAGETYP" in hdr else None)
            columns["instrument"].append(hdr["INSTRUME"] if "INSTRUME" in hdr else None)
            columns["site"].append(hdr["SITEID"] if "SITEID" in hdr else None)
            columns["telescope"].append(hdr["TELID"] if "TELID" in hdr else None)
            columns["rlevel"].append(hdr["RLEVEL"] if "RLEVEL" in hdr else None)

        # init df
        columns["filename"] = filenames
        self._data = pd.DataFrame(columns)

    def _filter_data(
        self,
        start: Optional[Time] = None,
        end: Optional[Time] = None,
        night: Optional[str] = None,
        site: Optional[str] = None,
        telescope: Optional[str] = None,
        instrument: Optional[str] = None,
        image_type: Optional[ImageType] = None,
        binning: Optional[str] = None,
        filter_name: Optional[str] = None,
        rlevel: Optional[int] = None,
    ) -> pd.DataFrame:
        """Filter data"""

        # filter
        data = self._data
        if start is not None:
            data = data[data["date-obs"] > start]
        if end is not None:
            data = data[data["date-obs"] < end]
        if night is not None:
            data = data[data["day-obs"] == night]
        if site is not None:
            data = data[data["site"] == site]
        if telescope is not None:
            data = data[data["telescope"] == telescope]
        if instrument is not None:
            data = data[data["instrument"] == instrument]
        if image_type is not None:
            data = data[data["image_type"] == image_type.value]
        if binning is not None:
            data = data[data["binning"] == binning]
        if filter_name is not None:
            data = data[data["filter"] == filter_name]
        if rlevel is not None:
            data = data[data["rlevel"] == rlevel]
        return data

    async def list_options(
        self,
        start: Optional[Time] = None,
        end: Optional[Time] = None,
        night: Optional[str] = None,
        site: Optional[str] = None,
        telescope: Optional[str] = None,
        instrument: Optional[str] = None,
        image_type: Optional[ImageType] = None,
        binning: Optional[str] = None,
        filter_name: Optional[str] = None,
        rlevel: Optional[int] = None,
    ) -> Dict[str, List[Any]]:
        """Returns a list of options restricted to the given parameters.

        Args:
            start: Start time for restriction.
            end: End time for restriction.
            night: Images in given night.
            site: From given site.
            telescope: From given telescope.
            instrument: From given instrument.
            image_type: With given image type.
            binning: With given binning.
            filter_name: With given filter.
            rlevel: In given reduction level.

        Returns:
            Dictionary with lists of "binnings", "filters", "imagetypes", "instruments", "sites", and "telescopes".
        """

        # filter
        data = self._filter_data(
            start, end, night, site, telescope, instrument, image_type, binning, filter_name, rlevel
        )

        # return results
        return {
            "binnings": data["binning"].unique(),
            "filters": data["filter"].unique(),
            "imagetypes": data["image_type"].unique(),
            "instruments": data["instrument"].unique(),
            "sites": data["site"].unique(),
            "telescopes": data["telescope"].unique(),
        }

    async def list_frames(
        self,
        start: Optional[Time] = None,
        end: Optional[Time] = None,
        night: Optional[str] = None,
        site: Optional[str] = None,
        telescope: Optional[str] = None,
        instrument: Optional[str] = None,
        image_type: Optional[ImageType] = None,
        binning: Optional[str] = None,
        filter_name: Optional[str] = None,
        rlevel: Optional[int] = None,
    ) -> List[FrameInfo]:
        """Returns a list of frames restricted to the given parameters.

        Args:
            start: Start time for restriction.
            end: End time for restriction.
            night: Images in given night.
            site: From given site.
            telescope: From given telescope.
            instrument: From given instrument.
            image_type: With given image type.
            binning: With given binning.
            filter_name: With given filter.
            rlevel: In given reduction level.

        Returns:
            List of frames.
        """

        # filter
        data = self._filter_data(
            start, end, night, site, telescope, instrument, image_type, binning, filter_name, rlevel
        )

        # create list of FrameInfos
        infos: List[FrameInfo] = []
        for _, row in data.iterrows():
            info = FrameInfo()
            info.id = row["filename"]
            info.filename = row["filename"]
            info.filter_name = row["filter"]
            info.binning = row["binning"]
            info.dateobs = row["date-obs"]
            infos.append(info)
        return infos

    async def download_frames(self, frames: List[FrameInfo]) -> List[Image]:
        """Download given frames.

        Args:
            frames: List of frames to download.

        Returns:
            List of Image objects.
        """

        # load frames
        images = []
        for frame in frames:
            images.append(Image.from_file(frame.filename))
        return images

    async def download_headers(self, infos: List[FrameInfo]) -> List[Dict[str, Any]]:
        """Download given headers.

        Args:
            infos: List of frames to download.

        Returns:
            List of dicts with headers.
        """

        # load frames
        headers = []
        for frame in infos:
            headers.append({k: v for k, v in fits.getheader(frame.filename).items()})
        return headers

    async def upload_frames(self, images: List[Image]) -> None:
        pass


__all__ = ["LocalArchive"]



from typing import Tuple
import astropy.units as u
from astropy.coordinates import SkyCoord, EarthLocation, AltAz, ICRS


def offset_altaz_to_radec(altaz: SkyCoord, dalt: float, daz: float) -> Tuple[float, float]:
    # convert to ra/dec
    p0 = altaz.transform_to(ICRS())
    p1 = altaz.spherical_offsets_by(daz * u.degree, dalt * u.degree).icrs

    # astropy hot-fix
    p0 = SkyCoord(ra=p0.ra, dec=p0.dec, frame="icrs")
    p1 = SkyCoord(ra=p1.ra, dec=p1.dec, frame="icrs")

    # offset
    dra, ddec = p0.spherical_offsets_to(p1)
    return float(dra.degree), float(ddec.degree)


def offset_radec_to_altaz(radec: SkyCoord, ra: float, dec: float, location: EarthLocation) -> Tuple[float, float]:
    # convert to alt/az
    altaz = AltAz(
        location=location,
        obstime=radec.obstime,
    )
    p0 = radec.transform_to(altaz)
    p1 = radec.spherical_offsets_by(ra * u.degree, dec * u.degree).transform_to(altaz)
    daz, dalt = p0.spherical_offsets_to(p1)
    return float(dalt.degree), float(daz.degree)



from __future__ import annotations

import asyncio
import functools
import inspect
import json
import logging
import re
import time
from collections.abc import Coroutine
from typing import Any, Callable, Dict, Type, List, Optional, TYPE_CHECKING, Tuple
import slixmpp
import slixmpp.exceptions
from slixmpp import ElementBase
from slixmpp.xmlstream import ET
import xml.sax.saxutils

from pyobs.comm import Comm
from pyobs.events import Event, LogEvent, ModuleOpenedEvent, ModuleClosedEvent
from pyobs.events.event import EventFactory
from pyobs.interfaces import Interface
from pyobs.utils import exceptions as exc
from .rpc import RPC
from .xmppclient import XmppClient

if TYPE_CHECKING:
    from pyobs.modules import Module

log = logging.getLogger(__name__)


class EventStanza(ElementBase):  # type: ignore
    name = "event"
    namespace = "pyobs:event"


class XmppComm(Comm):
    """A Comm class using XMPP.

    This Comm class uses an XMPP server (e.g. `ejabberd <https://www.ejabberd.im>`_) for communication between modules.
    Essentially required for a connection to the server is a JID, a JabberID. It can be specified in the configuration
    like this::

        comm:
            class: pyobs.comm.xmpp.XmppComm
            jid:  someuser@example.com/pyobs

    Using this, *pyobs* tries to connect to example.com as user ``someuser`` with resource ``pyobs``. Since ``pyobs``
    is the default resource, it can be omitted::

        jid:  someuser@example.com

    Alternatively, one can split the user, domain, and resource (if required) into three different parameters::

        user: someuser
        domain: example.com

    This comes in handy, if one wants to put the basic Comm configuration into a separate file. Imagine a ``_comm.yaml``
    in the same directory as the module config::

        comm_cfg: &comm
            class: pyobs.comm.sleekxmpp.XmppComm
            domain: example.com

    Now in the module configuration, one can simply do this::

        {include _comm.yaml}

        comm:
            <<: *comm
            user: someuser
            password: supersecret

    This allows for a super easy change of the domain for all configurations, which especially makes developing on
    different machines a lot easier.

    The ``server`` parameter can be used, when the server's hostname is different from the XMPP domain. This might,
    e.g., be the case, when connecting to a server via SSH port forwarding::

        jid:  someuser@example.com/pyobs
        server: localhost:52222

    Finally, always make sure that ``use_tls`` is set according to the server's settings, i.e. if it uses TLS, this
    parameter must be True, and False otherwise. Cryptic error messages will follow, if one does not set this properly.

    """

    __module__ = "pyobs.comm.xmpp"

    def __init__(
        self,
        jid: Optional[str] = None,
        user: Optional[str] = None,
        domain: Optional[str] = None,
        resource: str = "pyobs",
        password: str = "",
        server: Optional[str] = None,
        use_tls: bool = False,
        *args: Any,
        **kwargs: Any,
    ):
        """Create a new XMPP Comm module.

        Either a fill JID needs to be provided, or a set of user/domian/resource, from which a JID is built.

        Args:
            jid: JID to connect as.
            user: Username part of the JID.
            domain: Domain part of the JID.
            resource: Resource part of the JID.
            password: Password for given JID.
            server: Server to connect to. If not given, domain from JID is used.
            use_tls: Whether or not to use TLS.
        """
        Comm.__init__(self, *args, **kwargs)

        # variables
        self._connected = False
        self._online_clients: List[str] = []
        self._interface_cache: Dict[str, asyncio.Future[List[Type[Interface]]]] = {}
        self._user = user
        self._password = password
        self._domain = domain
        self._resource = resource
        self._server = server
        self._use_tls = use_tls
        self._loop = asyncio.get_event_loop()
        self._safe_send_attempts = 5
        self._safe_send_wait = 1
        self._module: Optional[Module] = None

        # build jid
        if jid:
            # resource given in jid?
            if "/" not in jid:
                jid += "/" + resource

            # get user/domain/resource and write it back to config
            m = re.match(r"([\w_\-\.]+)@([\w_\-\.]+)\/([\w_\-\.]+)", jid)
            if not m:
                log.error("Invalid JID format.")
                raise ValueError()
            self._user = m.group(1)
            self._domain = m.group(2)
            self._resource = m.group(3)

            # set jid itself
            self._jid = jid

        else:
            self._jid = "%s@%s/%s" % (self._user, self._domain, self._resource)

        #  client and RPC handler
        self._xmpp: Optional[XmppClient] = None
        self._rpc: Optional[RPC] = None

    def _set_module(self, module: Module) -> None:
        """Called, when the module connected to this Comm changes.

        Args:
            module: The module.
        """
        self._module = module

    async def open(self) -> None:
        """Open the connection to the XMPP server.

        Returns:
            Whether opening was successful.
        """

        # create client
        self._xmpp = XmppClient(self._jid, self._password)
        # self._xmpp = slixmpp.ClientXMPP(self._jid, password)
        self._xmpp.add_event_handler("pubsub_publish", self._handle_event)
        self._xmpp.add_event_handler("got_online", self._got_online)
        self._xmpp.add_event_handler("got_offline", self._got_offline)

        # server given?
        server = () if self._server is None else tuple(self._server.split(":"))

        # add features
        if self._module is not None:
            for i in self._module.interfaces:
                self._xmpp["xep_0030"].add_feature("pyobs:interface:%s" % i.__name__)

        # RPC
        self._rpc = RPC(self, self._xmpp, None)
        self._rpc.set_handler(self._module)

        # connect
        self._xmpp.connect(address=server, force_starttls=self._use_tls, disable_starttls=not self._use_tls)
        self._xmpp.init_plugins()

        # wait for connected
        if not await self._xmpp.wait_connect():
            if self.module is not None:
                self.module.quit()
            return

        # subscribe to events
        await self.register_event(LogEvent)

        # open Comm
        await Comm.open(self)

        # wait a little and finished
        await asyncio.sleep(1)
        self._connected = True

    async def close(self) -> None:
        """Close connection."""

        # close parent class
        await Comm.close(self)

        # disconnect from sleekxmpp server
        await self._xmpp.disconnect()

    @property
    def name(self) -> Optional[str]:
        """Name of this client."""
        return self._user

    def _failed_auth(self, event: Any) -> None:
        """Authentication failed.

        Args:
            event: XMPP event.
        """
        print("Authorization at server failed.")

    def _get_full_client_name(self, name: str) -> str:
        """Builds full JID from a given username.

        Args:
            name: Username to build JID for.

        Returns:
            Full JID for given user.
        """
        return name if "@" in name else "%s@%s/%s" % (name, self._domain, self._resource)

    async def get_interfaces(self, client: str) -> List[Type[Interface]]:
        """Returns list of interfaces for given client.

        Args:
            client: Name of client.

        Returns:
            List of supported interfaces.

        Raises:
            IndexError: If client cannot be found.
        """

        # full JID given?
        if "@" not in client:
            client = "%s@%s/%s" % (client, self._domain, self._resource)

        # return them from cache
        return await self._interface_cache[client]

    async def _get_interfaces(self, jid: str, attempts: int = 3) -> List[str]:
        """Return list of interfaces for the given JID.

        Args:
            jid: JID to get interfaces for.

        Returns:
            List of interface names or empty list, if an error occurred.
        """

        # request features
        try:
            info = await self._safe_send(self._xmpp["xep_0030"].get_info, jid=jid, cached=False)
        except (slixmpp.exceptions.IqError, slixmpp.exceptions.IqTimeout):
            return []

        # extract pyobs interface names
        if info is None:
            return []
        try:
            if isinstance(info, slixmpp.stanza.iq.Iq):
                info = info["disco_info"]
            prefix = "pyobs:interface:"
            interface_names = [i[len(prefix) :] for i in info["features"] if i.startswith(prefix)]
        except TypeError:
            raise IndexError()

        # IModule not in list?
        if "IModule" not in interface_names:
            # try again or quit?
            if attempts == 0:
                return []
            else:
                await asyncio.sleep(5)
                interface_names = await self._get_interfaces(jid, attempts - 1)

        # finished
        return interface_names

    async def _supports_interface(self, client: str, interface: Type[Interface]) -> bool:
        """Checks, whether the given client supports the given interface.

        Args:
            client: Client to check.
            interface: Interface to check.

        Returns:
            Whether or not interface is supported.
        """

        # full JID given?
        if "@" not in client:
            client = "%s@%s/%s" % (client, self._domain, self._resource)

        # update interface cache and get interface names
        interfaces = await self.get_interfaces(client)

        # supported?
        return interface in interfaces

    async def execute(self, client: str, method: str, annotation: Dict[str, Any], *args: Any) -> Any:
        """Execute a given method on a remote client.

        Args:
            client (str): ID of client.
            method (str): Method to call.
            annotation: Method annotation.
            *args: List of parameters for given method.

        Returns:
            Passes through return from method call.
        """

        # prepare
        if self._rpc is None:
            raise ValueError("No RPC.")
        jid = self._get_full_client_name(client)

        # call
        try:
            return await self._rpc.call(jid, method, annotation, *args)
        except slixmpp.exceptions.IqError:
            raise exc.RemoteError(client, f"Could not call {method} on {client}.")
        except slixmpp.exceptions.IqTimeout:
            raise exc.RemoteTimeoutError(client, f"Call to {method} on {client} timed out.")

    async def _got_online(self, msg: Any) -> None:
        """If a new client connects, add it to list.

        Args:
            msg: XMPP message.
        """

        # get jid, ignore event if it's myself
        jid = msg["from"].full
        if jid == self._jid:
            return

        # clear interface cache, just in case there is something there
        if jid in self._interface_cache:
            del self._interface_cache[jid]

        # create future for interfaces
        self._interface_cache[jid] = asyncio.get_running_loop().create_future()

        # request interfaces
        interface_names = await self._get_interfaces(jid)

        # if no interfaces are implemented (not even IModule), quit here
        if len(interface_names) == 0:
            module = jid[: jid.index["@"]]
            log.error(f"Module {module} does not seem to implement IModule, ignoring.")
            return

        # store interfaces
        self._interface_cache[jid].set_result(self._interface_names_to_classes(interface_names))

        # append to list
        if jid not in self._online_clients:
            self._online_clients.append(jid)

        # send event
        self._send_event_to_module(ModuleOpenedEvent(), msg["from"].username)

    def _got_offline(self, msg: Any) -> None:
        """If a new client disconnects, remove it from list.

        Args:
            msg: XMPP message.
        """

        # remove from list
        jid = msg["from"].full
        self._online_clients.remove(jid)

        # clear interface cache
        if jid in self._interface_cache:
            del self._interface_cache[jid]

        # send event
        self._send_event_to_module(ModuleClosedEvent(), msg["from"].username)

    @property
    def clients(self) -> List[str]:
        """Returns list of currently connected clients.

        Returns:
            (list) List of currently connected clients.
        """
        return [c[: c.find("@")] for c in self._online_clients]

    @property
    def client(self) -> XmppClient:
        """Returns the XMPP client.

        Returns:
            The XMPP client.
        """
        return self._xmpp

    async def send_event(self, event: Event) -> None:
        """Send an event to other clients.

        Args:
            event (Event): Event to send
        """

        # create stanza
        stanza = EventStanza()

        # dump event to JSON and escape it
        body = xml.sax.saxutils.escape(json.dumps(event.to_json()))

        # set xml and send event
        stanza.xml = ET.fromstring('<event xmlns="pyobs:event">%s</event>' % body)

        # send it
        await self._safe_send(
            self._xmpp["xep_0163"].publish,
            stanza,
            node="pyobs:event:%s" % event.__class__.__name__,
            callback=functools.partial(self._send_event_callback, event=event),
        )

        # send it to local module
        self._send_event_to_module(event, self._module.name)

    @staticmethod
    def _send_event_callback(iq: Any, event: Optional[Event] = None) -> None:
        """Called when an event has been successfully sent.

        Args:
            iq: Response package.
            event: Sent event.
        """
        log.debug("%s successfully sent.", event.__class__.__name__)

    async def _register_events(
        self, events: List[Type[Event]], handler: Optional[Callable[[Event, str], Coroutine[Any, Any, bool]]] = None
    ) -> None:
        # loop events
        for ev in events:
            # register event at XMPP
            self._xmpp["xep_0030"].add_feature("pyobs:event:%s" % ev.__name__)

            # if we have a handler, we're also interested in receiving such events
            if handler:
                # add interest
                self._xmpp["xep_0163"].add_interest("pyobs:event:%s" % ev.__name__)

        # update caps and send presence
        await self._safe_send(self._xmpp["xep_0115"].update_caps)
        self._xmpp.send_presence()

    async def _handle_event(self, msg: Any) -> None:
        """Handles an event.

        Args:
            msg: Received XMPP message.
        """

        # get body, unescape it, parse it
        # node = msg['pubsub_event'][items']['node']
        body = json.loads(xml.sax.saxutils.unescape(msg["pubsub_event"]["items"]["item"]["payload"].text))

        # do we have a <delay> element?
        delay = msg.xml.findall("{urn:sleekxmpp:delay}delay")
        if len(delay) > 0:
            # ignore this message
            return

        # did we send this?
        if msg["from"] == self._xmpp.boundjid.bare:
            return

        # create event and check timestamp
        event = EventFactory.from_dict(body)
        if event is None:
            return
        if time.time() - event.timestamp > 30:
            # event is more than 30 seconds old, ignore it
            # we do this do avoid resent events after a reconnect
            return

        # send it to module
        self._send_event_to_module(event, msg["from"].username)

    async def _safe_send(self, method: Callable[[Any], Coroutine[Any, Any, None]], *args: Any, **kwargs: Any) -> Any:
        """Safely send an XMPP message.

        Args:
            method: Method to call.
            *args: Parameters for method.
            **kwargs: Parameters for method.        except slixmpp.exceptions.IqError:
            raise RemoteException("Could not send command.")
        except slixmpp.exceptions.IqTimeout:
            raise exc()


        Returns:
            Return value from method.
        """

        # try multiple times
        for i in range(self._safe_send_attempts):
            try:
                # execute method and return result
                return await method(*args, **kwargs)

            except slixmpp.exceptions.IqTimeout:
                # timeout occurred, try again after some wait
                await asyncio.sleep(self._safe_send_wait)

        # never should reach this
        raise slixmpp.exceptions.IqTimeout

    def cast_to_simple_pre(self, value: Any, annotation: Optional[Any] = None) -> Tuple[bool, Any]:
        """Special treatment of single parameters when converting them to be sent via Comm.

        Args:
            value: Value to be treated.
            annotation: Annotation for value.

        Returns:
            A tuple containing a tuple that indicates whether this value should be further processed and a new value.
        """

        if type(value) == str:
            return True, xml.sax.saxutils.escape(value)
        else:
            return False, value

    def cast_to_real_post(self, value: Any, annotation: Optional[Any] = None) -> Tuple[bool, Any]:
        """Special treatment of single parameters when converting them after being sent via Comm.

        Args:
            value: Value to be treated.
            annotation: Annotation for value.

        Returns:
            A tuple containing a tuple that indicates whether this value should be further processed and a new value.
        """

        if type(value) == str:
            return True, xml.sax.saxutils.unescape(value)
        else:
            return False, value


__all__ = ["XmppComm"]



import functools
import inspect
from inspect import BoundArguments, Parameter
from enum import Enum
from typing import Any, get_args, Callable, Tuple, Optional, Type, Dict, get_origin


def iterate_params(
    value: Any,
    type_hint: Optional[Type[Any]] = None,
    method: Optional[Callable[[Any, Any], Tuple[bool, Optional[Any]]]] = None,
) -> Any:
    """Iterate values and type_hints and call a given method.

    Args:
        value: value to iterate.
        type_hint: type_hint for value.
        method: Method to run on each value. Should take one value and type_hint and return a bool indicating whether
            iteration should stop here and a new value.

    Returns:
        Same structure as input, but converted by method.
    """

    # call provided method
    if method:
        stop_iter, value = method(value, type_hint)
        if stop_iter:
            return value

    # okay, iterate value
    if value is None or type_hint is None or type_hint == Parameter.empty:
        # no response or no type_hint at all or Any
        return value

    elif isinstance(value, tuple):
        # handle tuple
        if type_hint and get_origin(type_hint) == tuple:
            return tuple(iterate_params(v, a, method) for v, a in zip(value, get_args(type_hint)))
        else:
            return tuple(iterate_params(v, None, method) for v in value)

    elif isinstance(value, list):
        # handle lists
        if type_hint and get_origin(type_hint) == list:
            typ = get_args(type_hint)[0]
            return [iterate_params(v, typ, method) for v in value]
        else:
            return [iterate_params(v, None, method) for v in value]

    elif isinstance(value, dict):
        # handle dict
        if type_hint and get_origin(type_hint) == dict:
            annk, annv = get_args(type_hint)
            return {iterate_params(k, annk, method): iterate_params(v, annv, method) for k, v in value.items()}
        else:
            return {iterate_params(k, None, method): iterate_params(v, None, method) for k, v in value.items()}

    else:
        # just return it, maybe cast to type_hint
        try:
            return type_hint(value) if type_hint else value
        except TypeError:
            return value


def cast_bound_arguments_to_simple(
    bound_arguments: BoundArguments,
    type_hints: Dict[str, Type[Any]],
    pre: Optional[Callable[[Any, Any], Tuple[bool, Optional[Any]]]] = None,
    post: Optional[Callable[[Any, Any], Tuple[bool, Optional[Any]]]] = None,
) -> None:
    """Cast the requested parameters, which are of simple types, to the types required by the method.

    Args:
        bound_arguments: Incoming parameters.
        type_hints: Type hints for parameters.
        pre: Method to call for each parameter before automatic handling.
        post: Method to call for each parameter after automatic handling.
    """

    # loop all arguments and cast
    cast_to_simple = functools.partial(__cast_to_simple, pre=pre, post=post)
    for key, value in bound_arguments.arguments.items():
        if key in ["self", "args", "kwargs"]:
            continue
        bound_arguments.arguments[key] = iterate_params(value, type_hints[key], cast_to_simple)


def cast_bound_arguments_to_real(
    bound_arguments: BoundArguments,
    type_hints: Dict[str, Type[Any]],
    pre: Optional[Callable[[Any, Any], Tuple[bool, Optional[Any]]]] = None,
    post: Optional[Callable[[Any, Any], Tuple[bool, Optional[Any]]]] = None,
) -> None:
    """Cast the requested parameters to real types.

    Args:
        bound_arguments: Incoming parameters.
        type_hints: Type hints for parameters.
        pre: Method to call for each parameter before automatic handling.
        post: Method to call for each parameter after automatic handling.
    """

    # loop all arguments and cast
    cast_to_real = functools.partial(__cast_to_real, pre=pre, post=post)
    for key, value in bound_arguments.arguments.items():
        bound_arguments.arguments[key] = iterate_params(value, type_hints[key], cast_to_real)


def cast_response_to_simple(
    value: Any,
    type_hint: Type[Any],
    pre: Optional[Callable[[Any, Any], Tuple[bool, Optional[Any]]]] = None,
    post: Optional[Callable[[Any, Any], Tuple[bool, Optional[Any]]]] = None,
) -> Any:
    """Cast a response from simple to the method's real types.

    Args:
        value: Response of method call.
        type_hint: type_hint for return value.
        pre: Method to call for each parameter before automatic handling.
        post: Method to call for each parameter after automatic handling.

    Returns:
        Same as input response, but with only simple types.
    """

    # cast
    cast_to_simple = functools.partial(__cast_to_simple, pre=pre, post=post)
    return iterate_params(value, type_hint, cast_to_simple)


def cast_response_to_real(
    value: Any,
    type_hint: Type[Any],
    pre: Optional[Callable[[Any, Any], Tuple[bool, Optional[Any]]]] = None,
    post: Optional[Callable[[Any, Any], Tuple[bool, Optional[Any]]]] = None,
) -> Any:
    """Cast a response from simple to the method's real types.

    Args:
        value: Response of method call.
        type_hint: type_hint of return value.
        pre: Method to call for each parameter before automatic handling.
        post: Method to call for each parameter after automatic handling.

    Returns:
        Same as input response, but with only simple types.
    """

    # cast
    cast_to_real = functools.partial(__cast_to_real, pre=pre, post=post)
    return iterate_params(value, type_hint, cast_to_real)


def __cast_to_simple(
    value: Any,
    type_hint: Type[Any],
    pre: Optional[Callable[[Any, Any], Tuple[bool, Optional[Any]]]] = None,
    post: Optional[Callable[[Any, Any], Tuple[bool, Optional[Any]]]] = None,
) -> Tuple[bool, Any]:
    """Default method for pre method that casts to simple types like str, int, etc.

    Args:
        value: Value to cast.
        type_hint: Type hint for cast.
        pre: Method to run before checking.
        post: Method to run after checking.

    Returns:
        Tuple of a boolean indicating whether to stop evaluating value and new value.
    """

    # init
    stop_iter = False

    # call provided pre-method
    if pre:
        stop_iter, value = pre(value, type_hint)
        if stop_iter:
            return True, value

    if inspect.isclass(type_hint) and issubclass(type_hint, Enum):
        # get string name for Enums
        stop_iter, value = True, value.value

    # call provided post-method
    if post:
        stop_iter_post, value = post(value, type_hint)
        if stop_iter_post:
            return True, value

    # stop iteration?
    return stop_iter, value


def __cast_to_real(
    value: Any,
    type_hint: Type[Any],
    pre: Optional[Callable[[Any, Any], Tuple[bool, Optional[Any]]]] = None,
    post: Optional[Callable[[Any, Any], Tuple[bool, Optional[Any]]]] = None,
) -> Tuple[bool, Any]:
    """Default method for pre method that casts to "real" Python types.

    Args:
        value: Value to cast.
        type_hint: Type hint for cast.
        pre: Method to run before checking.
        post: Method to run after checking.

    Returns:
        Tuple of a boolean indicating whether to stop evaluating value and new value.
    """

    # init
    stop_iter = False

    # call provided pre-method
    if pre:
        stop_iter, value = pre(value, type_hint)
        if stop_iter:
            return True, value

    if inspect.isclass(type_hint) and issubclass(type_hint, Enum):
        # get Enum from string
        stop_iter, value = True, type_hint(value)

    # call provided post-method
    if post:
        stop_iter_post, value = post(value, type_hint)
        if stop_iter_post:
            return True, value

    # stop iteration?
    return stop_iter, value



from __future__ import annotations
import contextlib
import time
import asyncio
from asyncio import Task
from collections.abc import Coroutine
from typing import Optional, List, Any, Union, Dict, TYPE_CHECKING

from pyobs.utils.types import cast_response_to_real

if TYPE_CHECKING:
    from pyobs.comm import Comm


async def event_wait(evt: asyncio.Event, timeout: float = 1.0) -> bool:
    # suppress TimeoutError because we'll return False in case of timeout
    with contextlib.suppress(asyncio.TimeoutError):
        await asyncio.wait_for(evt.wait(), timeout)
    return evt.is_set()


async def acquire_lock(lock: asyncio.Lock, timeout: float = 1.0) -> bool:
    # suppress TimeoutError because we'll return False in case of timeout
    try:
        await asyncio.wait_for(lock.acquire(), timeout)
        return True
    except asyncio.TimeoutError:
        return False


class Future(asyncio.Future):
    def __init__(
        self,
        empty: bool = False,
        annotation: Optional[Dict[str, Any]] = None,
        comm: Optional[Comm] = None,
        *args: Any,
        **kwargs: Any,
    ):
        asyncio.Future.__init__(self, *args, **kwargs)

        """Init new base future."""
        self.timeout: Optional[float] = None
        self.annotation = annotation
        self.comm = comm

        # already set?
        if empty:
            # fire event
            self.set_result(None)

    def set_timeout(self, timeout: float) -> None:
        """
        Sets a new timeout for the method call.
        """
        self.timeout = timeout

    def get_timeout(self) -> Optional[float]:
        """
        Returns async timeout.
        """
        return self.timeout

    def _wait_for_time(self, timeout: float = 0) -> None:
        """Waits a little.

        Args:
            time: Time to wait in seconds.
        """
        start = time.time()
        while not self.done() or time.time() - start > timeout:
            return
        raise TimeoutError

    def __await__(self) -> Any:
        # not finished? need to wait.
        if not self.done():
            try:
                # wait some 10s first
                self._wait_for_time(10)

            except TimeoutError:
                # got an additional timeout?
                if self.timeout is not None and self.timeout > 10:
                    # we already waited 10s, so subtract it
                    self._wait_for_time(self.timeout - 10.0)

        # not done? yield!
        if not self.done():
            self._asyncio_future_blocking = True
            yield self  # This tells Task to wait for completion.

        # still not done? raise exception.
        if not self.done():
            raise RuntimeError("await wasn't used with future")

        # get result
        result = self.result()

        # all ok, return value
        if self.annotation and self.comm:
            result = cast_response_to_real(
                result, self.annotation["return"], self.comm.cast_to_real_pre, self.comm.cast_to_real_post
            )
        return result

    @staticmethod
    async def wait_all(futures: List[Optional[Union[Future, Coroutine[Any, Any, Any], Task[Any]]]]) -> List[Any]:
        return [await fut for fut in futures if fut is not None]


__all__ = ["Future", "event_wait", "acquire_lock"]



from abc import ABCMeta

from .IExposure import IExposure
from .IData import IData


class ICamera(IData, IExposure, metaclass=ABCMeta):
    """The module controls a camera."""

    __module__ = "pyobs.interfaces"
    pass


__all__ = ["ICamera"]



from typing import Any, Dict, List, cast, Tuple, Optional
from urllib.parse import urljoin

import aiohttp

from pyobs.utils.time import Time


class Portal:
    def __init__(self, url: str, token: str):
        self.url = url
        self.token = token

    async def _get(self, path: str, timeout: int = 10, params: Optional[Dict[str, Any]] = None) -> Any:
        """Do a GET request on the portal.

        Args:
            url: URL to request.

        Returns:
            Response for request.

        Raises:
            RuntimeError if the call failed.
            TimeoutError if the call timed out.
        """

        # build header
        headers = {"Authorization": "Token " + self.token}

        async with aiohttp.ClientSession() as session:
            async with session.get(
                urljoin(self.url, path), headers=headers, timeout=timeout, params=params
            ) as response:
                if response.status != 200:
                    raise RuntimeError("Invalid response from portal: " + await response.text())
                return await response.json()

    async def last_changed(self) -> Time:
        t = await self._get("/api/last_changed/")
        return Time(t["last_change_time"])

    async def last_scheduled(self) -> Time:
        t = await self._get("/api/last_scheduled/")
        return Time(t["last_schedule_time"])

    async def schedulable_requests(self) -> List[Dict[str, Any]]:
        req = await self._get("/api/requestgroups/schedulable_requests/")
        return cast(List[Dict[str, Any]], req)

    async def proposals(self) -> List[Dict[str, Any]]:
        # init
        proposal_list: List[Dict[str, Any]] = []
        offset, limit = 0, 100

        # get everything!
        while True:
            # get batch of proposals
            proposals_new, count = await self._proposals(offset, limit)

            # empty set?
            if len(proposals_new) == 0:
                raise ValueError("Could not fetch data.")

            # add to list and increase offset
            proposal_list.extend(proposals_new)
            offset += limit

            # finished?
            if len(proposal_list) == count:
                return proposal_list

    async def _proposals(self, offset: int, limit: int) -> Tuple[List[Dict[str, Any]], int]:
        req = await self._get("/api/proposals/", params={"offset": offset, "limit": limit})
        return cast(List[Dict[str, Any]], req["results"]), req["count"]

    async def instruments(self) -> Dict[str, Any]:
        req = await self._get("/api/instruments/")
        return cast(Dict[str, Any], req)



from abc import ABCMeta

from .IData import IData
from .IExposure import IExposure


class ISpectrograph(IData, IExposure, metaclass=ABCMeta):
    """The module controls a camera."""

    __module__ = "pyobs.interfaces"
    pass


__all__ = ["ISpectrograph"]



from abc import ABCMeta

from .IStartStop import IStartStop
from .IExposureTime import IExposureTime


class IAutoGuiding(IStartStop, IExposureTime, metaclass=ABCMeta):
    """The module can perform auto-guiding."""

    __module__ = "pyobs.interfaces"


__all__ = ["IAutoGuiding"]
